# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional

from .crypto import EncryptKey, SignKey
from .identifier import ID
from .meta import Meta
from .profile import Profile


class EntityDataSource(ABC):
    """This interface is for getting information for entity(user/group)

        Entity Data Source
        ~~~~~~~~~~~~~~~~~~

        1. meta for user, which is generated by the user's private key
        2. meta for group, which is generated by the founder's private key
        3. meta key, which can verify message sent by this user(or group founder)
        4. profile key, which can encrypt message for the receiver(user)
    """

    @abstractmethod
    def meta(self, identifier: ID) -> Optional[Meta]:
        """
        Get meta for entity ID

        :param identifier: entity ID
        :return:           meta info
        """
        raise NotImplemented

    @abstractmethod
    def profile(self, identifier: ID) -> Optional[Profile]:
        """
        Get profile for entity ID

        :param identifier: entity ID
        :return:           profile info
        """
        pass


class UserDataSource(EntityDataSource):
    """This interface is for getting information for user

        User Data Source
        ~~~~~~~~~~~~~~~~

        (Encryption/decryption)
        1. public key for encryption
           if profile.key not exists, means it is the same key with meta.key
        2. private keys for decryption
           the private keys paired with [profile.key, meta.key]

        (Signature/Verification)
        3. private key for signature
           the private key paired with meta.key
        4. public keys for verification
           [meta.key]
    """

    @abstractmethod
    def contacts(self, identifier: ID) -> Optional[list]:
        """
        Get user's contacts list

        :param identifier: user ID
        :return: contact ID list
        """
        pass

    @abstractmethod
    def public_key_for_encryption(self, identifier: ID) -> Optional[EncryptKey]:
        """
        Get user's public key for encryption
        (profile.key or meta.key)

        :param identifier: user ID
        :return: public key
        """
        pass

    @abstractmethod
    def private_keys_for_decryption(self, identifier: ID) -> Optional[list]:
        """
        Get user's private keys for decryption
        (which paired with [profile.key, meta.key])

        :param identifier: user ID
        :return: private keys
        """
        raise NotImplemented

    @abstractmethod
    def private_key_for_signature(self, identifier: ID) -> Optional[SignKey]:
        """
        Get user's private key for signature
        (which paired with profile.key or meta.key)

        :param identifier: user ID
        :return: private key
        """
        raise NotImplemented

    @abstractmethod
    def public_keys_for_verification(self, identifier: ID) -> Optional[list]:
        """
        Get user's public keys for verification
        [profile.key, meta.key]

        :param identifier: user ID
        :return: public keys
        """
        pass


class GroupDataSource(EntityDataSource):
    """This interface is for getting information for group

        Group Data Source
        ~~~~~~~~~~~~~~~~~

        1. founder has the same public key with the group's meta.key
        2. owner and members should be set complying with the consensus algorithm
    """

    @abstractmethod
    def founder(self, identifier: ID) -> Optional[ID]:
        """
        Get founder of the group

        :param identifier: group ID
        :return: founder ID
        """
        pass

    @abstractmethod
    def owner(self, identifier: ID) -> Optional[ID]:
        """
        Get current owner of the group

        :param identifier: group ID
        :return: owner ID
        """
        pass

    @abstractmethod
    def members(self, identifier: ID) -> Optional[list]:
        """
        Get all members in the group

        :param identifier: group ID
        :return: member ID list
        """
        pass



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Union

from .crypto.utils import sha256, ripemd160, base58_encode, base58_decode
from .types import NetworkID


class Address(ABC):
    """This class is used to build address for ID

        Address for MKM ID
        ~~~~~~~~~~~~~~~~~~

        properties:
            network - address type
            number  - search number
    """

    def __new__(cls, address: str):
        """
        Create address object with string

        :param address: address string
        :return: Address object
        """
        if address is None:
            return None
        assert cls is Address, 'call Address() directly'
        if isinstance(address, Address):
            # return Address object directly
            return address
        # Address for broadcast
        length = len(address)
        # anywhere
        if length == len(ANYWHERE) and address.lower() == ANYWHERE:
            return ANYWHERE
        # everywhere
        if length == len(EVERYWHERE) and address.lower() == EVERYWHERE:
            return EVERYWHERE
        # try to create address object
        for clazz in cls.__address_classes:
            inst = clazz.__new__(clazz, address)
            if inst is not None:
                return inst
        raise ValueError('unrecognized address: %s' % address)

    @property
    @abstractmethod
    def network(self) -> int:
        """
        Get network type of address

        :return: integer as NetworkID
        """
        raise NotImplemented

    @property
    @abstractmethod
    def number(self) -> int:
        """
        Get search number of address

        :return: search number [0, 2^32)
        """
        raise NotImplemented

    @property
    def is_broadcast(self) -> bool:
        assert self.network is not None, 'address error: %s' % self
        if self.network == EVERYWHERE.network:
            # group address
            return self == EVERYWHERE
        elif self.network == ANYWHERE.network:
            # user address
            return self == ANYWHERE

    @property
    def is_user(self) -> bool:
        assert self.network is not None, 'address error: %s' % self
        return NetworkID.is_user(network=self.network)

    @property
    def is_group(self) -> bool:
        assert self.network is not None, 'address error: %s' % self
        return NetworkID.is_group(network=self.network)

    #
    #   Runtime
    #
    __address_classes = []  # class list

    @classmethod
    def register(cls, address_class) -> bool:
        """
        Register address class

        :param address_class: class for parsing ID.address
        :return: False on error
        """
        if issubclass(address_class, Address):
            cls.__address_classes.append(address_class)
        else:
            raise TypeError('%s must be subclass of Address' % address_class)
        return True


"""
    Address like BitCoin
    ~~~~~~~~~~~~~~~~~~~~

    data format: "network+digest+check_code"
        network    --  1 byte
        digest     -- 20 bytes
        check_code --  4 bytes

    algorithm:
        fingerprint = sign(seed, SK);
        digest      = ripemd160(sha256(fingerprint));
        check_code  = sha256(sha256(network + digest)).prefix(4);
        address     = base58_encode(network + digest + check_code);
"""


def check_code(data: bytes) -> bytes:
    # check code in BTC address
    return sha256(sha256(data))[:4]


def user_number(code: bytes) -> int:
    # get user number, which for remembering and searching user
    return int.from_bytes(code, byteorder='little')


class DefaultAddress(str, Address):

    def __init__(self, address: str):
        if self is address:
            # no need to init again
            return
        super().__init__()
        # get fields from string
        prefix_digest_code = base58_decode(address)
        if len(prefix_digest_code) != 25:
            raise ValueError('BTC address length error: %s' % address)
        # split them
        prefix = prefix_digest_code[:1]
        digest = prefix_digest_code[1:-4]
        code = prefix_digest_code[-4:]
        # check them
        if check_code(prefix + digest) != code:
            raise ValueError('BTC address check code error: %s' % address)
        network = ord(prefix)
        self.__network = network
        self.__number = user_number(code)

    @property
    def network(self) -> int:
        return self.__network

    @property
    def number(self) -> int:
        return self.__number

    #
    #   Factory
    #
    @classmethod
    def new(cls, data: bytes, network: Union[NetworkID, int]=0) -> Address:
        """
        Generate address with fingerprint and network ID

        :param data:    fingerprint (signature/key.data)
        :param network: address type
        :return:        Address object
        """
        if isinstance(network, NetworkID):
            network = network.value
        prefix = chr(network).encode('latin1')
        digest = ripemd160(sha256(data))
        code = check_code(prefix + digest)
        address = base58_encode(prefix + digest + code)
        return cls(address)


# register default address class
Address.register(DefaultAddress)  # default


"""
    Address for broadcast
    ~~~~~~~~~~~~~~~~~~~~~
"""


class ConstantAddress(str, Address):

    def __new__(cls, address: str, network: Union[NetworkID, int], number: int):
        self = super().__new__(cls, address)
        if isinstance(network, NetworkID):
            network = network.value
        self.__network = network
        self.__number = number
        return self

    @property
    def network(self) -> int:
        return self.__network

    @property
    def number(self) -> int:
        return self.__number


ANYWHERE = ConstantAddress(address="anywhere", network=NetworkID.Main, number=9527)
EVERYWHERE = ConstantAddress(address="everywhere", network=NetworkID.Group, number=9527)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Union

from .crypto.utils import base64_encode, base64_decode
from .crypto import PublicKey, PrivateKey, EncryptKey

from .types import NetworkID, MetaVersion
from .address import Address, DefaultAddress
from .identifier import ID


class Meta(dict, ABC):
    """This class is used to generate entity ID

        User/Group Meta data
        ~~~~~~~~~~~~~~~~~~~~

        data format: {
            version: 1,          // meta version
            seed: "moKy",        // user/group name
            key: "{public key}", // PK = secp256k1(SK);
            fingerprint: "..."   // CT = sign(seed, SK);
        }

        algorithm:
            fingerprint = sign(seed, SK);
    """

    # noinspection PyTypeChecker
    def __new__(cls, meta: dict):
        """
        Create meta for entity

        :param meta: meta info
        :return: Meta object
        """
        if meta is None:
            return None
        elif cls is Meta:
            if isinstance(meta, Meta):
                # return Meta object directly
                return meta
            # get class by meta version
            version = meta['version']
            if isinstance(version, MetaVersion):
                version = version.value
            else:
                version = int(version)
            clazz = cls.__meta_classes.get(version)
            if clazz is None:
                raise ModuleNotFoundError('meta version not supported: %s' % meta)
            return clazz.__new__(clazz, meta)
        # subclass
        return super().__new__(cls, meta)

    def __init__(self, meta: dict):
        if self is meta:
            # no need to init again
            return
        super().__init__(meta)
        # Meta algorithm version
        version = meta['version']
        if isinstance(version, MetaVersion):
            version = version.value
        else:
            version = int(version)
        self.__version = version
        # lazy
        self.__key: Union[PublicKey, EncryptKey, None] = None
        self.__seed: str = None
        self.__fingerprint: bytes = None
        self.__status: int = 0  # 1 for valid, -1 for invalid

    def __eq__(self, other) -> bool:
        """ Check whether they can generate same IDs """
        if not isinstance(other, dict):
            return False
        if super().__eq__(other):
            return True
        # check with ID
        other = Meta(other)
        identifier = other.generate_identifier(network=NetworkID.Main)
        return self.match_identifier(identifier)

    @property
    def version(self) -> int:
        """
        Meta algorithm version

            0x01 - username@address
            0x02 - btc_address
            0x03 - username@btc_address
            0x04 - eth_address
            0x05 - username@eth_address
            ...
        """
        return self.__version

    @property
    def has_seed(self) -> bool:
        return MetaVersion.has_seed(version=self.version)

    @property
    def key(self) -> Union[PublicKey, EncryptKey]:
        """
        Public key (used for signature)

            RSA
            ECC

        :return: public key
        """
        if self.__key is None:
            self.__key = PublicKey(self['key'])
        return self.__key

    @property
    def seed(self) -> Optional[str]:
        """
        Seed to generate fingerprint

        :return: ID.name
        """
        if self.__seed is None and self.has_seed:
            # MKM, ExBTC, ExETH, ...
            self.__seed = self['seed']
        return self.__seed

    @property
    def fingerprint(self) -> Optional[bytes]:
        """
        Fingerprint to verify ID and public key

            Build: fingerprint = sign(seed, privateKey)
            Check: verify(seed, fingerprint, publicKey)

        :return: signature
        """
        if self.__fingerprint is None and self.has_seed:
            # MKM, ExBTC, ExETH, ...
            self.__fingerprint = base64_decode(self['fingerprint'])
        return self.__fingerprint

    @property
    def valid(self) -> bool:
        """
        Check meta valid
        (must call this when received a new meta from network)

        :return: True on valid
        """
        if self.__status is 0:
            key = self.key
            if key is None:
                # meta.key should not be empty
                self.__status = -1
            elif self.has_seed:
                seed = self.seed
                fingerprint = self.fingerprint
                if seed is None or fingerprint is None:
                    # seed and fingerprint should not be empty
                    self.__status = -1
                elif key.verify(data=seed.encode('utf-8'), signature=fingerprint):
                    # fingerprint matched
                    self.__status = 1
                else:
                    # fingerprint not matched
                    self.__status = -1
            else:
                # BTC, ETH, ...
                self.__status = 1
        return self.__status == 1

    def match(self, public_key: PublicKey=None, identifier: ID=None, address: Address=None):
        if public_key is not None:
            return self.match_public_key(public_key=public_key)
        elif identifier is not None:
            return self.match_identifier(identifier=identifier)
        elif address is not None:
            return self.match_address(address=address)

    def match_public_key(self, public_key: PublicKey) -> bool:
        """
        Check whether match public key

        :param public_key: user's public key
        :return: True on matched
        """
        if not self.valid:
            return False
        if self.key == public_key:
            return True
        if self.has_seed:
            # check whether keys equal by verifying signature
            seed = self.seed
            fingerprint = self.fingerprint
            return public_key.verify(data=seed.encode('utf-8'), signature=fingerprint)
        else:
            # NOTICE: ID with BTC/ETH address has no username, so
            #         just compare the key.data to check matching
            return False

    def match_identifier(self, identifier: ID) -> bool:
        """
        Check ID(name+address) with meta info
        (must call this when received a new meta from network)

        :param identifier: user ID
        :return: True on matched
        """
        if identifier is None:
            return False
        return self.generate_identifier(identifier.address.network) == identifier

    def match_address(self, address: Address) -> bool:
        """
        Check address with meta info

        :param address: user's ID.address
        :return: True on matched
        """
        if address is None:
            return False
        return self.generate_address(network=address.network) == address

    def generate_identifier(self, network: Union[NetworkID, int]) -> ID:
        """
        Generate ID with meta info and network ID

        :param network: ID type
        :return: ID object
        """
        if isinstance(network, NetworkID):
            network = network.value
        address = self.generate_address(network=network)
        if address is not None:
            return ID.new(name=self.seed, address=address)

    @abstractmethod
    def generate_address(self, network: int) -> Address:
        """
        Generate address with meta info and network ID

        :param network: address type
        :return: Address object
        """
        # NOTICE: must check meta valid before generate address
        raise NotImplemented

    #
    #  Factories
    #
    @classmethod
    def new(cls, key: PublicKey, seed: str=None, fingerprint: bytes=None, version: MetaVersion=MetaVersion.Default):
        """
        Create new meta info

        :param version:     - meta version
        :param key:         - public key
        :param seed:        - user/group name
        :param fingerprint: - signature
        :return: Meta object
        """
        if isinstance(version, MetaVersion):
            version = version.value
        if MetaVersion.has_seed(version=version):
            # MKM, ExBTC, ExETH, ...
            meta = {
                'version': version,
                'seed': seed,
                'key': key,
                'fingerprint': base64_encode(fingerprint),
            }
        else:
            # BTC, ETH, ...
            meta = {
                'version': version,
                'key': key,
            }
        # new Meta(dict)
        return cls(meta)

    @classmethod
    def generate(cls, private_key: PrivateKey, seed: str='', version: MetaVersion=MetaVersion.Default):
        """
        Generate meta info with seed and private key

        :param private_key: - user/founder private key
        :param seed:        - user/group name
        :param version:     - meta version
        :return:
        """
        if isinstance(version, MetaVersion):
            version = version.value
        if MetaVersion.has_seed(version=version):
            # MKM, ExBTC, ExETH, ...
            # generate fingerprint with private key
            fingerprint = private_key.sign(seed.encode('utf-8'))
            dictionary = {
                'version': version,
                'seed': seed,
                'key': private_key.public_key,
                'fingerprint': base64_encode(fingerprint),
            }
        else:
            # BTC, ETH, ...
            dictionary = {
                'version': version,
                'key': private_key.public_key,
            }
        # new Meta(dict)
        return cls(dictionary)

    #
    #   Runtime
    #
    __meta_classes = {}  # int -> class

    @classmethod
    def register(cls, version: Union[MetaVersion, int], meta_class=None) -> bool:
        """
        Register meta class with version

        :param version:  meta version for different meta algorithms
        :param meta_class: if content class is None, then remove with type
        :return: False on error
        """
        if isinstance(version, MetaVersion):
            version = version.value
        else:
            version = int(version)
        if meta_class is None:
            cls.__meta_classes.pop(version, None)
        elif issubclass(meta_class, Meta):
            cls.__meta_classes[version] = meta_class
        else:
            raise TypeError('%s must be subclass of Meta' % meta_class)
        return True


"""
    Default Meta for generate ID with address
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    version:
        0x01 - MKM
        
    algorithm:
        CT      = fingerprint;
        hash    = ripemd160(sha256(CT));
        code    = sha256(sha256(network + hash)).prefix(4);
        address = base58_encode(network + hash + code);
        number  = u_int(code);
"""


class DefaultMeta(Meta):

    def __init__(self, meta: dict):
        if self is meta:
            # no need to init again
            return
        super().__init__(meta)
        # id caches
        self.__ids: dict = {}  # int -> ID

    def generate_identifier(self, network: int) -> ID:
        # check cache
        identifier = self.__ids.get(network)
        if identifier is None:
            # generate and cache it
            identifier = super().generate_identifier(network=network)
            assert identifier.valid, 'failed to generate ID with network: %s' % network
            self.__ids[network] = identifier
        return identifier

    def generate_address(self, network: int) -> Address:
        assert self.version == MetaVersion.MKM, 'meta version error: %d' % self.version
        assert self.valid, 'meta not valid: %s' % self
        # check cache
        identifier: ID = self.__ids.get(network)
        if identifier is not None:
            return identifier.address
        # generate
        return DefaultAddress.new(data=self.fingerprint, network=network)


# register meta class with version
Meta.register(version=MetaVersion.MKM, meta_class=DefaultMeta)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Group for assembly
    ~~~~~~~~~~~~~~~~~~

    Group with members
"""

from typing import Optional

from .identifier import ID
from .entity import Entity
from .delegate import GroupDataSource


class Group(Entity):
    """This class is for creating group

        Group for organizing users
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

            roles:
                founder
                owner
                members
                administrators - Optional
                assistants     - Optional
    """

    def __init__(self, identifier: ID):
        super().__init__(identifier=identifier)
        # once the group founder is set, it will never change
        self.__founder: ID = None

    @Entity.delegate.getter
    def delegate(self) -> Optional[GroupDataSource]:
        facebook = super().delegate
        assert facebook is None or isinstance(facebook, GroupDataSource), 'error: %s' % facebook
        return facebook

    # @delegate.setter
    # def delegate(self, value: GroupDataSource):
    #     super(Group, Group).delegate.__set__(self, value)

    @property
    def founder(self) -> Optional[ID]:
        if self.__founder is None:
            self.__founder = self.delegate.founder(identifier=self.identifier)
        return self.__founder

    @property
    def owner(self) -> Optional[ID]:
        return self.delegate.owner(identifier=self.identifier)

    @property
    def members(self) -> Optional[list]:
        return self.delegate.members(identifier=self.identifier)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any

from .crypto import VerifyKey, SignKey, EncryptKey


class TAI(ABC):
    """
        The Additional Information
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        'Meta' is the information for entity which never changed,
            which contains the key for verify signature;
        'TAI' is the variable part,
            which could contain a public key for asymmetric encryption.
    """

    # noinspection PyTypeChecker
    def __new__(cls, profile: dict):
        if profile is None:
            return None
        assert cls is TAI, 'call TAI() directly'
        if isinstance(profile, TAI):
            # return Profile object directly
            return profile
        # try to create Profile object
        for clazz in cls.__tai_classes:
            inst = clazz.__new__(clazz, profile)
            if inst is not None:
                return inst

    @property
    @abstractmethod
    def identifier(self) -> str:
        """
        Entity ID

        :return: Entity ID
        """
        raise NotImplemented

    @property
    @abstractmethod
    def valid(self) -> bool:
        """
        If signature matched

        :return: True on valid
        """
        raise NotImplemented

    @property
    def key(self) -> Optional[EncryptKey]:
        """
        Get public key to encrypt message for user

        :return: public key
        """
        return None

    """
        Properties
        ~~~~~~~~~~
        
        Inner dictionary which will be serialized to 'data'
    """

    @property
    @abstractmethod
    def _properties(self) -> Optional[dict]:
        """
        Get all properties when valid

        :return: inner dictionary
        """
        raise NotImplemented

    def get_property(self, key: str) -> Optional[Any]:
        """
        Get property value with key

        :param key: property key
        :return: property value
        """
        properties = self._properties
        if properties is None:
            return None
        return properties.get(key)

    @abstractmethod
    def set_property(self, key: str, value: Any=None):
        """
        Update property with key and data
        (this will clear 'data' and 'signature')

        :param key:   property key
        :param value: property value
        """
        properties = self._properties
        assert properties is not None, 'failed to get properties'
        if value is None:
            properties.pop(key, None)
        else:
            properties[key] = value

    """
        Sign/Verify profile data
        ~~~~~~~~~~~~~~~~~~~~~~~~
    """

    @abstractmethod
    def verify(self, public_key: VerifyKey) -> bool:
        """
        Verify 'data' and 'signature' with public key

        :param public_key: public key in meta.key
        :return: True on signature matched
        """
        raise NotImplemented

    @abstractmethod
    def sign(self, private_key: SignKey) -> bytes:
        """
        Encode properties to 'data' and sign it to 'signature'

        :param private_key: private key match meta.key
        :return: signature
        """
        raise NotImplemented

    #
    #   Runtime
    #
    __tai_classes = []  # class list

    @classmethod
    def register(cls, tai_class) -> bool:
        """
        Register TAI class

        :param tai_class: class for parsing Profile
        :return: False on error
        """
        from .profile import Profile
        assert tai_class is not TAI, 'should not register TAI itself!'
        assert tai_class is not Profile, 'cannot register Profile!'
        if issubclass(tai_class, TAI):
            cls.__tai_classes.append(tai_class)
        else:
            raise TypeError('%s must be subclass of TAI' % tai_class)
        return True



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import json
from typing import Optional, Union, Any

from .crypto.utils import base64_decode, base64_encode
from .crypto import PublicKey, EncryptKey, VerifyKey, SignKey
from .identifier import ID
from .tai import TAI


class Profile(dict, TAI):

    # noinspection PyTypeChecker
    def __new__(cls, profile: dict):
        if profile is None:
            return None
        elif cls is Profile:
            inst = TAI.__new__(TAI, profile)
            if inst is not None:
                # created by subclass
                return inst
        # new Profile(dict)
        return super().__new__(cls, profile)

    def __init__(self, profile: dict):
        if self is profile:
            # no need to init again
            return
        super().__init__(profile)
        # lazy
        self.__identifier: str = None
        self.__data: bytes = None
        self.__signature: bytes = None
        self.__properties: dict = None
        self.__status: int = 0  # 1 for valid, -1 for invalid

    @property
    def identifier(self) -> str:
        """ Profile ID """
        if self.__identifier is None:
            self.__identifier = self['ID']
        return self.__identifier

    @property
    def _data(self) -> Optional[bytes]:
        """ Profile properties data """
        if self.__data is None:
            string: str = self.get('data')
            if string is not None:
                self.__data = string.encode('utf-8')
        return self.__data

    @property
    def _signature(self) -> Optional[bytes]:
        """ Profile properties signature """
        if self.__signature is None:
            base64: str = self.get('signature')
            if base64 is not None:
                self.__signature = base64_decode(base64)
        return self.__signature

    @property
    def valid(self) -> bool:
        return self.__status >= 0

    """
        Profile Properties
        ~~~~~~~~~~~~~~~~~~
    """

    @property
    def _properties(self) -> Optional[dict]:
        """ Load properties from data """
        if self.__status < 0:
            # invalid
            return None
        if self.__properties is None:
            data = self._data
            if data is None:
                # create new properties
                self.__properties = {}
            else:
                # get properties from data
                self.__properties = json.loads(data)
                assert isinstance(self.__properties, dict)
        return self.__properties

    def set_property(self, key: str, value: Any=None):
        """ Update profile property with key and data """
        # 1. reset status
        self.__status = 0
        # 2. update property value with name
        super().set_property(key=key, value=value)
        # 3. clear data signature after properties changed
        self.pop('data', None)
        self.pop('signature', None)
        self.__data = None
        self.__signature = None

    """
        Name
        ~~~~
        Nickname for user
        Title for group
    """

    @property
    def name(self) -> Optional[str]:
        value = self.get_property(key='name')
        if value is not None:
            return value

    @name.setter
    def name(self, value: str):
        self.set_property(key='name', value=value)

    """
        Sign/Verify profile data
        ~~~~~~~~~~~~~~~~~~~~~~~~
    """

    def verify(self, public_key: VerifyKey) -> bool:
        """
        Verify 'data' and 'signature' with public key

        :param public_key: public key in meta.key
        :return: True on signature matched
        """
        if self.__status > 0:
            # already verify OK
            return True
        data = self._data
        signature = self._signature
        if data is None:
            # NOTICE: if data is empty, signature should be empty at the same time
            #         this happen while profile not found
            if signature is None:
                self.__status = 0
            else:
                # data signature error
                self.__status = -1
        elif signature is None:
            # signature error
            self.__status = -1
        elif public_key.verify(data, signature):
            # signature matched
            self.__status = 1
        # NOTICE: if status is 0, it doesn't mean the profile is invalid,
        #         try another key
        return self.__status == 1

    def sign(self, private_key: SignKey) -> bytes:
        """
        Encode properties to 'data' and sign it to 'signature'

        :param private_key: private key match meta.key
        :return: signature
        """
        if self.__status > 0:
            # already signed
            return self.__signature
        self.__status = 1
        data: str = json.dumps(self._properties)
        self.__data = data.encode('utf-8')
        self.__signature = private_key.sign(self.__data)
        self['data'] = data  # JsON string
        self['signature'] = base64_encode(self.__signature)
        return self.__signature

    @classmethod
    def new(cls, identifier: ID):
        """ Create new empty profile object with entity ID """
        profile = {
            'ID': identifier,
        }
        return cls(profile)


class UserProfile(Profile):

    # noinspection PyTypeChecker
    def __new__(cls, profile: dict):
        if profile is None:
            return None
        elif cls is UserProfile:
            # 1. if ID type is user, convert to UserProfile
            # 2. if public key not exists, no need to convert to UserProfile
            identifier = profile.get('ID')
            if isinstance(identifier, ID):
                if not identifier.is_user:
                    return None
            elif 'avatar' not in profile and 'key' not in profile:
                # not a user profile
                return None
        # new UserProfile(dict)
        return super().__new__(cls, profile)

    def __init__(self, profile: dict):
        if self is profile:
            # no need to init again
            return
        super().__init__(profile)
        # lazy
        self.__key: PublicKey = None  # EncryptKey

    """
        Public Key for encryption
        ~~~~~~~~~~~~~~~~~~~~~~~~~
        For safety considerations, the profile.key which used to encrypt message data
        should be different with meta.key
    """
    @property
    def key(self) -> Union[PublicKey, EncryptKey, None]:
        if self.__key is None:
            key = self.get_property(key='key')
            if key is not None:
                self.__key = PublicKey(key=key)
        return self.__key

    @key.setter
    def key(self, value: Union[PublicKey, EncryptKey]):
        self.__key = value
        self.set_property('key', value)

    """
        Nickname
        ~~~~~~~~
    """
    @property
    def name(self) -> Optional[str]:
        value = self.get_property(key='name')
        if value is not None:
            return value
        # get from 'names'
        array = self.get_property(key='names')
        if array is not None and len(array) > 0:
            return array[0]

    @name.setter
    def name(self, value: str):
        self.set_property(key='name', value=value)

    """
        Avatar
        ~~~~~~
    """
    @property
    def avatar(self) -> Optional[str]:
        value = self.get_property(key='avatar')
        if value is not None:
            return value
        # get from 'names'
        array = self.get_property(key='photos')
        if array is not None and len(array) > 0:
            return array[0]

    @avatar.setter
    def avatar(self, value: str):
        self.set_property(key='avatar', value=value)


# register user profile class
Profile.register(UserProfile)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Immortals
    ~~~~~~~~~

    Built-in users
"""

import os
from typing import Optional

from .dos import JSONFile
from .crypto import PrivateKey, EncryptKey, DecryptKey, SignKey
from .identifier import ID
from .meta import Meta
from .profile import Profile
from .user import User
from .delegate import UserDataSource


def load_resource_file(filename: str) -> dict:
    directory = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(directory, 'res', filename)
    file = JSONFile(path=path)
    return file.read()


class Immortals(UserDataSource):

    # Immortal Hulk (195-183-9394)
    HULK = 'hulk@4YeVEN3aUnvC1DNUufCq1bs9zoBSJTzVEj'
    # Monkey King (184-083-9527)
    MOKI = 'moki@4WDfe3zZ4T7opFSi3iDAKiuTnUHjxmXekk'

    def __init__(self):
        super().__init__()
        # caches
        self.__ids = {}
        self.__private_keys = {}
        self.__metas = {}
        self.__profiles = {}
        self.__users = {}
        # load built-in users
        self.__load_user(identifier=ID(self.HULK))
        self.__load_user(identifier=ID(self.MOKI))

    def __load_user(self, identifier: ID):
        assert identifier.valid, 'ID error: %s' % identifier
        self.cache_id(identifier=identifier)
        # load meta for ID
        meta = self.__load_meta(filename=identifier.name+'_meta.js')
        self.cache_meta(meta=meta, identifier=identifier)
        # load private key for ID
        key = self.__load_private_key(filename=identifier.name+'_secret.js')
        self.cache_private_key(private_key=key, identifier=identifier)
        # load profile for ID
        profile = self.__load_profile(filename=identifier.name + '_profile.js')
        self.cache_profile(profile=profile, identifier=identifier)

    @staticmethod
    def __load_meta(filename: str) -> Optional[Meta]:
        return Meta(load_resource_file(filename=filename))

    @staticmethod
    def __load_private_key(filename: str) -> Optional[PrivateKey]:
        return PrivateKey(load_resource_file(filename=filename))

    def __load_profile(self, filename: str) -> Optional[Profile]:
        profile = Profile(load_resource_file(filename=filename))
        assert profile is not None, 'failed to load profile: %s' % filename
        # copy 'name'
        name = profile.get('name')
        if name is None:
            names = profile.get('names')
            if names is not None and len(names) > 0:
                profile.set_property('name', names[0])
        else:
            profile.set_property('name', name)
        # copy 'avatar'
        avatar = profile.get('avatar')
        if avatar is None:
            photos = profile.get('photos')
            if photos is not None and len(photos) > 0:
                profile.set_property('avatar', photos[0])
        else:
            profile.set_property('avatar', avatar)
        # sign
        self.__sign_profile(profile=profile)
        return profile

    def __sign_profile(self, profile: Profile) -> bytes:
        identifier = self.identifier(profile.identifier)
        key = self.private_key_for_signature(identifier)
        assert key is not None, 'failed to get private key for signature: %s' % identifier
        return profile.sign(private_key=key)

    def cache_id(self, identifier: ID) -> bool:
        assert identifier.valid, 'ID not valid: %s' % identifier
        self.__ids[identifier] = identifier
        return True

    def cache_meta(self, meta: Meta, identifier: ID) -> bool:
        assert meta.match_identifier(identifier), 'meta not match: %s, %s' % (identifier, meta)
        self.__metas[identifier] = meta
        return True

    def cache_private_key(self, private_key: PrivateKey, identifier: ID) -> bool:
        assert self.meta(identifier).key.match(private_key), 'private key error: %s, %s' % (identifier, private_key)
        self.__private_keys[identifier] = private_key
        return True

    def cache_profile(self, profile: Profile, identifier: ID) -> bool:
        assert profile.valid, 'profile not valid: %s' % profile
        assert identifier == profile.identifier, 'profile not match: %s, %s' % (identifier, profile)
        self.__profiles[profile.identifier] = profile
        return True

    def cache_user(self, user: User) -> bool:
        if user.delegate is None:
            user.delegate = self
        self.__users[user.identifier] = user
        return True

    # ----

    def identifier(self, string: str) -> Optional[ID]:
        if string is None:
            return None
        elif isinstance(string, ID):
            return string
        assert isinstance(string, str), 'ID string error: %s' % string
        return self.__ids.get(string)

    def user(self, identifier: ID) -> Optional[User]:
        assert identifier.is_user, 'ID type error: %s' % identifier
        user = self.__users.get(identifier)
        if user is None:
            # only create exists account
            if identifier in self.__ids:
                user = User(identifier=identifier)
                self.cache_user(user=user)
        return user

    #
    #   EntityDataSource
    #
    def meta(self, identifier: ID) -> Optional[Meta]:
        return self.__metas.get(identifier)

    def profile(self, identifier: ID) -> Optional[Profile]:
        return self.__profiles.get(identifier)

    #
    #   UserDataSource
    #
    def contacts(self, identifier: ID) -> Optional[list]:
        assert identifier.is_user, 'ID error: %s' % identifier
        if identifier not in self.__ids:
            return None
        array = []
        for key, value in self.__ids.items():
            if key == identifier:
                continue
            array.append(value)
        return array

    def public_key_for_encryption(self, identifier: ID) -> Optional[EncryptKey]:
        assert identifier.is_user, 'ID error: %s' % identifier
        # NOTICE: return nothing to use profile.key or meta.key
        return None

    def private_keys_for_decryption(self, identifier: ID) -> Optional[list]:
        assert identifier.is_user, 'ID error: %s' % identifier
        key = self.__private_keys.get(identifier)
        if isinstance(key, DecryptKey):
            return [key]

    def private_key_for_signature(self, identifier: ID) -> Optional[SignKey]:
        assert identifier.is_user, 'ID error: %s' % identifier
        return self.__private_keys.get(identifier)

    def public_keys_for_verification(self, identifier: ID) -> Optional[list]:
        assert identifier.is_user, 'ID error: %s' % identifier
        # NOTICE: return nothing to use meta.key
        return None



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

from .crypto import EncryptKey, DecryptKey, SignKey, VerifyKey
from .entity import Entity
from .delegate import UserDataSource


class User(Entity):
    """This class is for creating user

        User for communication
        ~~~~~~~~~~~~~~~~~~~~~~

        functions:
            (User)
            1. verify(data, signature) - verify (encrypted content) data and signature
            2. encrypt(data)           - encrypt (symmetric key) data
            (LocalUser)
            3. sign(data)    - calculate signature of (encrypted content) data
            4. decrypt(data) - decrypt (symmetric key) data
    """

    @Entity.delegate.getter
    def delegate(self) -> Optional[UserDataSource]:
        facebook = super().delegate
        assert facebook is None or isinstance(facebook, UserDataSource), 'error: %s' % facebook
        return facebook

    # @delegate.setter
    # def delegate(self, value: UserDataSource):
    #     super(User, User).delegate.__set__(self, value)

    @property
    def contacts(self) -> Optional[list]:
        """
        Get all contacts of the user

        :return: contacts list
        """
        return self.delegate.contacts(identifier=self.identifier)

    def __meta_key(self) -> VerifyKey:
        meta = self.meta
        assert meta is not None, 'failed to get meta for user: %s' % self.identifier
        return meta.key

    def __profile_key(self) -> Optional[EncryptKey]:
        profile = self.profile
        if profile is None or not profile.valid:
            # profile not found or not valid
            return None
        return profile.key

    # NOTICE: meta.key will never changed, so use profile.key to encrypt
    #         is the better way
    def __encrypt_key(self) -> EncryptKey:
        # 0. get key from data source
        key = self.delegate.public_key_for_encryption(identifier=self.identifier)
        if key is not None:
            return key
        # 1. get key from profile
        key = self.__profile_key()
        if key is not None:
            return key
        # 2. get key from meta
        key = self.__meta_key()
        if isinstance(key, EncryptKey):
            return key
        raise LookupError('failed to get encrypt key for user: %s' % self.identifier)

    # NOTICE: I suggest using the private key paired with meta.key to sign message
    #         so here should return the meta.key
    def __verify_keys(self) -> list:
        # 0. get keys from data source
        keys = self.delegate.public_keys_for_verification(identifier=self.identifier)
        if keys is not None and len(keys) > 0:
            return keys
        keys = []
        # # 1. get key from profile
        # key = self.__profile_key()
        # if isinstance(key, VerifyKey):
        #     keys.append(key)
        # 2. get key from meta
        key = self.__meta_key()
        assert key is not None, 'meta.key not found: %s' % self.identifier
        keys.append(key)
        return keys

    def verify(self, data: bytes, signature: bytes) -> bool:
        """
        Verify data and signature with user's public keys

        :param data:
        :param signature:
        :return:
        """
        keys = self.__verify_keys()
        for key in keys:
            if key.verify(data=data, signature=signature):
                # matched!
                return True

    def encrypt(self, data: bytes) -> bytes:
        """Encrypt data, try profile.key first, if not found, use meta.key

        :param data: plaintext
        :return: ciphertext
        """
        key = self.__encrypt_key()
        assert key is not None, 'failed to get encrypt key for user: %s' % self.identifier
        return key.encrypt(data=data)

    #
    #   interfaces for local user
    #

    # NOTICE: I suggest use the private key which paired to meta.key
    #         to sign message
    def __sign_key(self) -> SignKey:
        return self.delegate.private_key_for_signature(identifier=self.identifier)

    # NOTICE: if you provide a public key in profile for encryption
    #         here you should return the private key paired with profile.key
    def __decrypt_keys(self) -> list:
        return self.delegate.private_keys_for_decryption(identifier=self.identifier)

    def sign(self, data: bytes) -> bytes:
        """
        Sign data with user's private key

        :param data: message data
        :return: signature
        """
        key = self.__sign_key()
        assert key is not None, 'failed to get sign key for user: %s' % self.identifier
        return key.sign(data=data)

    def decrypt(self, data: bytes) -> Optional[bytes]:
        """
        Decrypt data with user's private key(s)

        :param data: ciphertext
        :return: plaintext
        """
        keys = self.__decrypt_keys()
        assert keys is not None and len(keys) > 0, 'failed to get decrypt keys: %s' % self.identifier
        for key in keys:
            assert isinstance(key, DecryptKey), 'decrypt key error: %s' % key
            try:
                # try decrypting it with each private key
                plaintext = key.decrypt(data=data)
                if plaintext is not None:
                    # OK!
                    return plaintext
            except ValueError:
                # this key not match, try next one
                continue
